// お習字ロボットアーム

package org.microbridge.servocontrol;

import java.io.IOException;
import java.util.ArrayList;
import org.microbridge.server.Server;
import android.app.Activity;
import android.graphics.Color;
import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.view.Window;
import android.view.WindowManager;
import android.widget.Button;

// お習字ロボットアームアプリ本体
public class RobotArm extends Activity
{
	// microbridgeサーバ(TCPサーバ)
	private Server server;
	// 一画ぶんの運筆データ (座標データの配列)
	private ArrayList<InvKinema> stroke;
	// 全筆順の運筆データ (座標データの配列の配列)
	private ArrayList<ArrayList<InvKinema>> strokeArray;
	// 初期ポジションの座標データ
	private InvKinema POS0;
	// 角度[deg]によるサーボ制御コマンド (解像度が低いのでもう使わない)
	private final byte COMMAND_DEGREE = 0x5A;
	// パルス幅[us]によるサーボ制御コマンド
	private final byte COMMAND_USEC = 0x55;
	// 記録/停止ボタン
	private Button buttonRec;
	// 再生ボタン
	private Button buttonPlay;
	// お習字画面
	private BrushView brushView;
	// 各サーボの角度
	private float[] axes;
	// 記録中か？
	public boolean isRecording;

	// アプリの初期化
	@Override
	public void onCreate(Bundle savedInstanceState)
	{
		super.onCreate(savedInstanceState);
		
		// 二の腕の長さ[mm]
		InvKinema.L = 100;
		// 手首から先の長さ[mm]
		InvKinema.d = 30;
		// お習字の物理領域のオフセット[mm]
		InvKinema.Y0 = 230;
		InvKinema.X0 = 0;
		// お習字の物理領域のサイズ[mm]
		InvKinema.width = 66;
		InvKinema.height = 100;
		// お習字の画面領域のサイズ[pixel]
		InvKinema.sWidth = 320;
		InvKinema.sHeight = 480;

		// サーボの初期位置はすべてのサーボを90度
		POS0 = new InvKinema(90,70,60,60);
		axes = new float[4];
		isRecording = false;

		// microbridgeサーバ(TCPサーバ)を生成
		server = null;
		try
		{
			// ポート番号 4567 (Arduino側と合わせる)
			server = new Server(4567);
			server.start();
		} catch (IOException e)
		{
			Log.e("microbridge", "Unable to start TCP server", e);
			System.exit(-1);
		}

		// Set full screen view, no title
		getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN, WindowManager.LayoutParams.FLAG_FULLSCREEN);
		requestWindowFeature(Window.FEATURE_NO_TITLE);
		setContentView(R.layout.main);

		// お習字画面
		brushView = (BrushView)findViewById(R.id.brushView1);
		brushView.setParent(this);

		// 記録/停止ボタン
		buttonRec = (Button)findViewById(R.id.button1);
		buttonRec.setOnClickListener(new View.OnClickListener() {
			public void onClick(View v) {
				// 記録ボタンを押したとき
				if(isRecording == false){
					// 停止ボタンに切り替える
					buttonRec.setTextColor(Color.BLACK);
					buttonRec.setText("STOP");
					// 再生ボタンは無効
					buttonPlay.setEnabled(false);
					// 運筆データをクリア
					brushView.clearPath();
					strokeArray = new ArrayList<ArrayList<InvKinema>>();
					isRecording = true;
				// 停止ボタンを押したとき
				}else{
					// 記録ボタンに切り替える
					buttonRec.setTextColor(Color.RED);
					buttonRec.setText("REC");
					// 再生ボタンは有効
					buttonPlay.setEnabled(true);
					isRecording = false;
				}
			}
		});

		// 再生ボタン
		buttonPlay = (Button)findViewById(R.id.button2);
		buttonPlay.setEnabled(false);
		buttonPlay.setOnClickListener(new View.OnClickListener() {
			// 再生ボタンを押したとき
			public void onClick(View v) {
				// 再生(ロボットアームに書かせる)
				play();
			}
		});

		// 筆を初期位置に移動
		setAxes(POS0);
	}

	// 指が触れたとき(一画の始点)
	void touchDown(double sx, double sy)
	{
		// 新しい一画の始まり
		stroke = new ArrayList<InvKinema>();
		strokeArray.add(stroke);
		// 始点の座標を逆キネ計算して運筆データに追加
		touchMove(sx, sy);
	}
	
	// 指が離れたとき(一画の終点)
	void touchUp(double sx, double sy)
	{
		// 終点の座標を逆キネ計算して運筆データに追加
		touchMove(sx, sy);
	}
	
	// 指が動いたとき(一画の途中の点)
	void touchMove(double sx, double sy)
	{
		// 座標を逆キネ計算して運筆データに追加
		InvKinema ki = new InvKinema((int)sx,(int)sy);
		try{
			ki.calcAxes();
			stroke.add(ki);
		}catch(Exception e){
			errlog(e.getMessage());
		}
	}
	
	// 再生(ロボットアームに書かせる)
	void play()
	{
		try{
			// 筆を初期位置に移動
			setAxes(POS0);
			Thread.sleep(100);
			
			// 一画ずつ
			for(int i=0;i<strokeArray.size();i++)
			{
				ArrayList<InvKinema> st = strokeArray.get(i);
				
				// 一画の始点から終点まで
				for(int j=0;j<st.size();j++)
				{
					// 各点のデータを取得
					InvKinema ik = st.get(j);
					Log.i("RobotArm", "i="+i+" j="+j+" x="+ik.sx+" y="+ik.sy
							+" th="+ik.theta+" phi1="+ik.phi1+" phi2="+ik.phi2+" phi3="+ik.phi3);
					
					// 始点のときは、筆を上げて始点まで移動
					if(j==0){
						sweepAxes(ik.bringUp());
//						setAxes(ik.bringUp());
						Thread.sleep(100);
					}
					// この点まで筆を運ぶ
					sweepAxes(ik);
//					setAxes(ik);
					Thread.sleep(100);
					
					// 終点のときは、筆を上げる
					if(j==st.size()-1){
						sweepAxes(ik.bringUp());
//						setAxes(ik.bringUp());
						Thread.sleep(100);
					}
				}
			}
			// 筆を初期位置に戻す
			sweepAxes(POS0);
//			setAxes(POS0);

		}catch(InterruptedException e){
			errlog("Thread.sleep error.");
		}
	}
	
	// エラーログ
	void errlog(String msg)
	{
		Log.e("PenArm",msg);
	}

	// 指定の点まで筆を移動(一気に)
	void setAxes(InvKinema ik)
	{
		// 角度で指令(解像度が低いのでもう使わない)
//		axes[0] = ik.theta;
//		axes[1] = ik.phi1;
//		axes[2] = ik.phi2;
//		axes[3] = ik.phi3;
//		sendDeg2Arduino(axes);
		
		// パルス幅で指令
		// パルス幅[us] = 角度[deg] * 10 + 600us
		axes[0] = (float)(ik.theta *10.0 + 600.0);
		axes[1] = (float)(ik.phi1  *10.0 + 600.0);
		axes[2] = (float)(ik.phi2  *10.0 + 600.0);
		axes[3] = (float)(ik.phi3  *10.0 + 600.0);
		sendPw2Arduino(axes);
	}
	
	// 指定の点まで筆を移動(中割点を計算して経由)
	void sweepAxes(InvKinema ik)
	{
		float[] temp = new float[4];
		float[] dist = new float[4];
		
		// 角度(解像度が低いのでもう使わない)
//		dist[0] = ik.theta;
//		dist[1] = ik.phi1;
//		dist[2] = ik.phi2;
//		dist[3] = ik.phi3;
		
		// パルス幅[us] = 角度[deg] * 10 + 600us
		dist[0] = (float)(ik.theta *10.0 + 600.0);
		dist[1] = (float)(ik.phi1  *10.0 + 600.0);
		dist[2] = (float)(ik.phi2  *10.0 + 600.0);
		dist[3] = (float)(ik.phi3  *10.0 + 600.0);
		
		// なるべく細かく中割点の数を決める
		float d0 = Math.abs(dist[0] - axes[0]);
		float d1 = Math.abs(dist[1] - axes[1]);
		int nakawari = (int)((d0>d1) ? d0 : d1);
		if (nakawari<1) nakawari = 1;
		Log.i("Nakawari0","nakawari="+nakawari+" d0="+d0+" d1="+d1);
		
		// 各中割点から目的点まで順に
		for(int i=1;i<=nakawari;i++)
		{
			// 中割点を計算
			for(int j=0;j<4;j++){
				temp[j] = axes[j] + ((dist[j]-axes[j])*(float)i)/(float)nakawari;
			}
			Log.i("Nakawari", " th="+temp[0]+" phi1="+temp[1]+" phi2="+temp[2]+" phi3="+temp[3]);
			
			// Arduinoに指令を送る
//			sendDeg2Arduino(temp);
			sendPw2Arduino(temp);
			try {
//				Thread.sleep(100);
				Thread.sleep(10);
			} catch (InterruptedException e) {
				errlog("Thread.sleep error.");
			}
		}
		for(int j=0;j<4;j++){
			axes[j] = dist[j];
		}
	}
	
	// サーボの角度指令をArduinoに送る (解像度が低いのでもう使わない)
	void sendDeg2Arduino(float[] angle)
	{
		byte[] data = new byte[5];
		
		data[0] = COMMAND_DEGREE;
		data[1] = (byte)angle[0];
		data[2] = (byte)angle[1];
		data[3] = (byte)angle[2];
		data[4] = (byte)angle[3];
		
		try{
			server.send(data);
		}catch(IOException e){
			errlog("server.send error.");
		}
	}
	
	// サーボのパルス幅指令をArduinoに送る
	void sendPw2Arduino(float[] angle)
	{
		byte[] data = new byte[9];
		int[] t = new int[4];
		
		t[0] = (int)angle[0];
		t[1] = (int)angle[1]; 
		t[2] = (int)angle[2];
		t[3] = (int)angle[3];
		
		data[0] = COMMAND_USEC;
		data[1] = (byte)((t[0]>>8) & 0xFF);
		data[2] = (byte)( t[0]     & 0xFF);
		data[3] = (byte)((t[1]>>8) & 0xFF);
		data[4] = (byte)( t[1]     & 0xFF);
		data[5] = (byte)((t[2]>>8) & 0xFF);
		data[6] = (byte)( t[2]     & 0xFF);
		data[7] = (byte)((t[3]>>8) & 0xFF);
		data[8] = (byte)( t[3]     & 0xFF);
		Log.i("send",data[1]+" "+data[2]+" "+data[3]+" "+data[4]+" "+data[5]+" "+data[6]+" "+data[7] + " " + data[8]);
		
		try{
			server.send(data);
		}catch(IOException e){
			errlog("server.send error.");
		}
	}
}